@@ -3,6 +3,11 @@ from collections import defaultdict
 
 
 class Pdb:  # TODO optionally save as gro? & think of trajectories
+    prot_map = {'ALA': 'A', 'CYS': 'C', 'CYX': 'C', 'CYM': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G',
+                'HIS': 'H', 'HIE': 'H', 'HID': 'H', 'HSD': 'H', 'HSE': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L',
+                'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V',
+                'TRP': 'W', 'TYR': 'Y', "GLUP": "E", "ASPP": "D"}
+
     def __init__(self, filename=None, top=None, altloc='A', **kwargs):
         self.fname = filename
         if self.fname:
@@ -19,20 +24,36 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         if not self.atoms and self.top:
             self.atoms = [Atom.from_top_entry(entry) for mol in self.top.molecules for entry in mol.atoms]
         self.altloc = altloc
+        self.conect = {}  # TODO will require further adjustments
         self._atom_format = "ATOM  {:>5d} {:4s}{:1s}{:4s}{:1s}{:>4d}{:1s}   " \
                             "{:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:>2s}\n"
+        self._atom_format_gro = "{:>5d}{:5s}{:>5s}{:>5d}{:8.3f}{:8.3f}{:8.3f}\n"
         self._cryst_format = "CRYST1{:9.3f}{:9.3f}{:9.3f}{:7.2f}{:7.2f}{:7.2f} P 1           1\n"
 
     def __repr__(self):
         return "PDB file {} with {} atoms".format(self.fname, len(self.atoms))
 
     def add_top(self, top, **kwargs):
+        """
+        Adds a Top object to the current Pdb object, enabling some
+        operations that couple the two
+        :param top: str, path to the .top file
+        :param kwargs: dict, extra parameters to pass to Top constructor
+        :return: None
+        """
         self.top = gml.Top(top, **kwargs)
         if self.top and not self.top.pdb:
             self.top.pdb = self
 
     @classmethod
     def from_selection(cls, pdb, selection):
+        """
+        Creates
+        :param pdb:
+        :param selection:
+        :return:
+        """
+        # TODO what to do in case of a bound Top object? implement Top.slice()?
         selected_indices = pdb.select_atoms(selection)
         new_pdb = Pdb()
         new_pdb.atoms = [atom for n, atom in enumerate(pdb.atoms) if n in selected_indices]
@@ -57,38 +78,51 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         new_pdb.altloc = orig_pdb.altloc
         return new_pdb
 
-    def print_protein_sequence(self):
-        mapping = {'ALA': 'A', 'CYS': 'C', 'CYX': 'C', 'CYM': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G',
-                   'HIS': 'H', 'HIE': 'H', 'HID': 'H', 'HSD': 'H', 'HSE': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L',
-                   'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V',
-                   'TRP': 'W', 'TYR': 'Y', "GLUP": "E", "ASPP": "D"}
-        chains = list({a.chain for a in self.atoms if a.atomname == 'CA'})
-        if chains == [' ']:
+    def print_protein_sequence(self, gaps=False):
+        chains = list({a.chain.strip() for a in self.atoms if a.atomname == 'CA'})
+        sequences = []
+        if not chains:
             print("No chains in the molecule, run Pdb.add_chains() first")
             return
         for ch in sorted(chains):
             cas = set(self.select_atoms(f'name CA and chain {ch}'))
-            atoms = [a for n, a in enumerate(self.atoms) if n in cas]
-            print(''.join([mapping[i.resname] for i in atoms]))
+            atoms = [a for n, a in enumerate(self.atoms) if n in cas and a.altloc in [' ', self.altloc]]
+            if not gaps:
+                sequences.append(''.join([Pdb.prot_map[i.resname] if i.resname in Pdb.prot_map.keys() else 'X'
+                                          for i in atoms]))
+            else:
+                res = [a.resnum for n, a in enumerate(self.atoms) if n in cas and a.altloc in [' ', self.altloc]]
+                seq = []
+                counter = 0
+                for r in range(res[0], res[-1]+1):
+                    if r in res:
+                        rname = atoms[counter].resname
+                        counter += 1
+                        seq.append(Pdb.prot_map[rname] if rname in Pdb.prot_map.keys() else 'X')
+                    else:
+                        seq.append('-')
+                sequences.append(''.join(seq))
+        return sequences
 
     def print_nucleic_sequence(self):
         mapping = defaultdict(lambda: 'X')
+        sequences = []
         mapping.update({'DA': "A", 'DG': "G", 'DC': "C", 'DT': "T", 'DA5': "A", 'DG5': "G", 'DC5': "C", 'DT5': "T",
                         'DA3': "A", 'DG3': "G", 'DC3': "C", 'DT3': "T", 'RA': "A", 'RG': "G", 'RC': "C", 'RU': "U",
                         'RA5': "A", 'RG5': "G", 'RC5': "C", 'RU5': "U", 'RA3': "A", 'RG3': "G", 'RC3': "C", 'RU3': "U",
                         'A': "A", 'G': "G", 'C': "C", 'U': "U", 'A5': "A", 'G5': "G", 'C5': "C", 'U5': "U",
                         'A3': "A", 'G3': "G", 'C3': "C", 'U3': "U"})
-        chains = list({a.chain for a in self.atoms if a.atomname == "O4'"})
+        chains = list({a.chain.strip() for a in self.atoms if a.atomname == "O4'"})
+        if not chains:
+            print("No chains in the molecule, run Pdb.add_chains() first")
+            return
         for ch in sorted(chains):
             cas = set(self.select_atoms(f"name O4' and chain {ch}"))
             atoms = [a for n, a in enumerate(self.atoms) if n in cas]
-            print(''.join([mapping[i.resname] for i in atoms]))
+            sequences.append(''.join([mapping[i.resname] if i.resname in mapping.keys() else 'X' for i in atoms]))
+        return sequences
 
     def find_missing(self):
-        map_pro = {'ALA': 'A', 'CYS': 'C', 'CYX': 'C', 'CYM': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G',
-                   'HIS': 'H', 'HIE': 'H', 'HID': 'H', 'HSD': 'H', 'HSE': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L',
-                   'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V',
-                   'TRP': 'W', 'TYR': 'Y', "GLUP": "E", "ASPP": "D"}
         map_nuc = {'DA': "A", 'DG': "G", 'DC': "C", 'DT': "T", 'DA5': "A", 'DG5': "G", 'DC5': "C", 'DT5': "T",
                    'DA3': "A", 'DG3': "G", 'DC3': "C", 'DT3': "T", 'RA': "A", 'RG': "G", 'RC': "C", 'RU': "U",
                    'RA5': "A", 'RG5': "G", 'RC5': "C", 'RU5': "U", 'RA3': "A", 'RG3': "G", 'RC3': "C", 'RU3': "U",
@@ -109,22 +143,21 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         for at in self.atoms:
             if (at.resname, at.resnum) != curr_res:
                 if 'CA' in atomlist:
-                    full = set(pro_bb + pro_sc[map_pro[curr_res[0]]])
+                    full = set(pro_bb + pro_sc[Pdb.prot_map[curr_res[0]]])
                     if not full.issubset(set(atomlist)):
-                        if map_pro[curr_res[0]] in alt.keys():
-                            modfull = set([alt[map_pro[curr_res[0]]][1] if x == alt[map_pro[curr_res[0]]][0] else x
-                                           for x in full])
+                        if Pdb.prot_map[curr_res[0]] in alt.keys():
+                            modfull = set([alt[Pdb.prot_map[curr_res[0]]][1] if x == alt[Pdb.prot_map[curr_res[0]]][0]
+                                           else x for x in full])
                             if not modfull.issubset(set(atomlist)):
                                 print(f"atoms {modfull.difference(set(atomlist))} missing from residue {curr_res}")
                         else:
                             print(f"atoms {full.difference(set(atomlist))} missing from residue {curr_res}")
-
                 # TODO implement for nucleic
                 curr_res = (at.resname, at.resnum)
                 atomlist = []
             atomlist.append(at.atomname)
 
-    def add_chains(self, serials=None, chain=None, offset=0, maxwarn=100):
+    def add_chains(self, serials=None, chain=None, offset=0, maxwarn=100, cutoff=10, protein_only=False, nopbc=False):
         """
         Given a matching Top instance, adds chain identifiers to atoms
         based on the (previously verified) matching between invididual
@@ -137,19 +170,35 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         :param chain: chain to set for serials (default is use consecutive letters)
         :param offset: int, start chain ordering from letter other than A
         :param maxwarn: int, max number of warnings before an error shows up
+        :param cutoff: float, distance threshold (in A) for chain separation if using geometric criteria
+        :param protein_only: bool, whether to only add chains to protein residues
+        :param nopbc: bool, whether to ignore PBC distances (assumes molecule is whole)
         :return: None
         """
+        base_char = 65 + offset  # 65 is ASCII for "A"
+        curr_resid = None
+        prev_atom = None
         if not self.top:
-            ch = chain if chain is not None else 'X'
             for atom in self.atoms:
-                if atom.chain == ' ':
-                    atom.chain = ch
+                if not prev_atom:
+                    prev_atom = atom
+                    curr_resid = atom.resnum
+                if atom.resnum != curr_resid:
+                    dist = self._atoms_dist_pbc(atom, prev_atom, nopbc)
+                    prev_atom = atom
+                    curr_resid = atom.resnum
+                    if dist > cutoff:
+                        if (protein_only and atom.resname in Pdb.prot_map.keys()) or not protein_only:
+                            base_char += 1
+                if base_char > 90:
+                    break
+                if (protein_only and atom.resname in Pdb.prot_map.keys()) or not protein_only:
+                    atom.chain = chr(base_char)
         else:
             self.check_top(maxwarn=maxwarn)
             if (serials is None and chain is not None) or (serials is not None and chain is None):
                 raise ValueError("Both serials and chain have to be specified simultaneously")
             excluded = {'SOL', 'HOH', 'TIP3', 'K', 'NA', 'CL', 'POT'}
-            base_char = 65 + offset  # 65 is ASCII for "A"
             index = 0
             for mol_name in self.top.system.keys():
                 if mol_name.upper() not in excluded:
@@ -175,28 +224,23 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
             raise ValueError("a Top object has not been assigned; molecule info missing")
         index, err = 0, 0
         self._remove_altloc()
-        for mol_name in self.top.system.keys():
-            mol = self.top.get_molecule(mol_name)
-            n_mols = self.top.system[mol_name]
-            atom_subsection = mol.get_subsection('atoms')
-            atom_entries = [e for e in atom_subsection if isinstance(e, gml.EntryAtom)]
-            for m in range(n_mols):
-                for n, a in enumerate(atom_entries):
-                    try:
-                        rtrn = self._check_mismatch(atom_entries[n], self.atoms[index], mol_name)
-                    except IndexError:
-                        raise RuntimeError("Mismatch encountered: PDB has {} atoms while topology "
-                                           "has {}".format(n + 1, index + 1, len(self.atoms), self.top.natoms))
-                    if rtrn:
-                        if fix_pdb:
-                            self.atoms[index].atomname = atom_entries[n].atomname
-                        elif fix_top:
-                            atom_entries[n].atomname = self.atoms[index].atomname
-                    err += rtrn
-                    index += 1
-                    if err > maxwarn > -1:
-                        raise RuntimeError("Error: too many warnings; use maxwarn=N to allow for up to N exceptions,"
-                                           "or -1 to allow for any number of them")
+        for atom_top, atom_pdb in zip(self.top.atoms, self.atoms):
+            index += 1
+            try:
+                rtrn = self._check_mismatch(atom_top, atom_pdb, atom_top.molname)
+            except IndexError:
+                raise RuntimeError("Mismatch encountered: PDB has {} atoms while topology "
+                                   "has {}".format(len(self.atoms), len(self.top.atoms)))
+            if rtrn:
+                if fix_pdb:
+                    self.atoms[index].atomname = atom_top.atomname
+                elif fix_top:
+                    atom_top.atomname = self.atoms[index].atomname
+            err += rtrn
+            index += 1
+            if err > maxwarn > -1:
+                raise RuntimeError("Error: too many warnings; use maxwarn=N to allow for up to N exceptions,"
+                                   "or -1 to allow for any number of them")
         print("Check passed, all names match")
     
     @staticmethod
@@ -216,15 +260,56 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         """
         self.atoms = [a for a in self.atoms if a.altloc in [' ', self.altloc]]
     
-    def _write_atom(self, atom):
+    def _write_atom(self, atom, pdb=True):
         atom.serial %= 100000
         atom.resnum %= 10000
         if atom.occ > 1000:
             atom.occ %= 1000
         if atom.beta > 1000:
             atom.beta %= 1000
-        return self._atom_format.format(atom.serial, atom.atomname, atom.altloc, atom.resname, atom.chain, atom.resnum,
-                                        atom.insert, atom.x, atom.y, atom.z, atom.occ, atom.beta, atom.element)
+        if pdb:
+            return self._atom_format.format(atom.serial, atom.atomname, atom.altloc, atom.resname, atom.chain,
+                                            atom.resnum, atom.insert, atom.x, atom.y, atom.z, atom.occ, atom.beta,
+                                            atom.element)
+        else:
+            return self._atom_format_gro.format(atom.resnum, atom.resname, atom.atomname, atom.serial, atom.x, atom.y,
+                                                atom.z)
+
+    @staticmethod
+    def _write_conect(atom, bonded):
+        return 'CONECT' + ('{:>5}' * (len(bonded)+1)).format(atom, *bonded) + '\n'
+
+    def tip3_to_opc(self):
+        names = {"TIP3", "WAT", "SOL", "HOH", "TIP"}
+        o_names = {"O", "OW"}
+        h_names = {"H", "HW"}
+        new_atoms = []
+        water_mol = {}
+        for n, a in enumerate(self.atoms):
+            if a.resname in names:
+                if a.atomname in o_names:
+                    water_mol["O"] = a
+                if a.atomname[:-1] in h_names:
+                    try:
+                        _ = water_mol["H1"]
+                    except KeyError:
+                        water_mol["H1"] = a
+                    else:
+                        water_mol["H2"] = a
+                if len(water_mol) == 3:
+                    new_atoms.append(a)
+                    new_atom = Atom(self._write_atom(water_mol["O"]))
+                    new_atom.atomname = "MW"
+                    offs = 0.147722363
+                    new_atom.set_coords([a + offs*(b + c - 2*a) for a, b, c
+                                         in zip(new_atom.coords, water_mol["H1"].coords, water_mol["H2"].coords)])
+                    new_atoms.append(new_atom)
+                    water_mol = {}
+                else:
+                    new_atoms.append(a)
+            else:
+                new_atoms.append(a)
+        self.atoms = new_atoms
     
     def renumber_atoms(self):
         for n, atom in enumerate(self.atoms, 1):
@@ -251,7 +336,7 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
             vec = vector
         else:
             raise RuntimeError("In repositioning, please use either p1/p2 selections or specify the vector")
-        movable = self.atoms[self.select_atom(atomsel)]
+        movable = self.get_atom(atomsel)
         hook_xyz = self.get_coords()[self.select_atom(hooksel)]
         vec_len = sum([x**2 for x in vec])**0.5
         scale = bondlength/vec_len
@@ -326,6 +411,11 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         if renumber:
             self.renumber_atoms()
 
+    def match_elements(self):
+        for a in self.atoms:
+            if not a.element:
+                a.element = [x for x in a.atomname if not x.isdigit()][0]
+
     def select_atoms(self, selection_string):
         sel = gml.SelectionParser(self)
         return sel(selection_string)
@@ -338,6 +428,12 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         elif len(result) < 1:
             raise RuntimeError("Selection {} returned no atoms".format(selection_string, result))
         return result[0]
+
+    def get_atom(self, selection_string):
+        return self.atoms[self.select_atom(selection_string)]
+
+    def get_atoms(self, selection_string):
+        return [self.atoms[i] for i in self.select_atoms(selection_string)]
     
     def same_residue_as(self, query_iter):
         new_list = []
@@ -358,13 +454,20 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
     def _atoms_dist(at1, at2):
         return ((at2.x - at1.x)**2 + (at2.y - at1.y)**2 + (at2.z - at1.z)**2)**0.5
 
+    def _atoms_dist_pbc(self, at1, at2, nopbc=False):
+        if not self.box[3] == self.box[4] == self.box[5] == 90.0 and not nopbc:
+            raise RuntimeError("Only rectangular boxes supported for PBC-based distanes")
+        return (min([abs(at2.x - at1.x), self.box[0] - abs(at2.x - at1.x)]) ** 2 +
+                min([abs(at2.y - at1.y), self.box[1] - abs(at2.y - at1.y)]) ** 2 +
+                min([abs(at2.z - at1.z), self.box[2] - abs(at2.z - at1.z)]) ** 2) ** 0.5
+
     @staticmethod
     def _atoms_vec(at1, at2):
         return at2.x - at1.x, at2.y - at1.y, at2.z - at1.z
         
     @staticmethod
     def _parse_contents(contents):
-        atoms, remarks = [], []
+        atoms, remarks, conect = [], [], {}
         box = [7.5, 7.5, 7.5, 90, 90, 90]  # generic default, will be overwritten if present
         for line in contents:
             if line.startswith('ATOM') or line.startswith('HETATM'):
@@ -374,6 +477,11 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
                       [float(line[33+7*a:33+7*(a+1)]) for a in range(3)]
             elif not line.startswith('TER') and not line.startswith('END'):
                 remarks.append(line)
+            elif line.startswith('CONECT'):
+                conect[int(line[6:11].strip())] = []
+                for i in [11, 16, 21, 26]:
+                    if line[i:i+5].strip():
+                        conect[int(line[6:11].strip())].append(int(line[i:i+5].strip()))
             if line.startswith('END') or line.startswith('ENDMDL'):
                 break
         return atoms, tuple(box), remarks
@@ -381,7 +489,7 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
     def mutate_protein_residue(self, resid, target, chain=''):
         self.renumber_atoms()
         chstr = 'chain {} and '.format(chain) if chain else ''
-        orig = self.atoms[self.select_atom('{}resid {} and name CA'.format(chstr, resid))]
+        orig = self.get_atom('{}resid {} and name CA'.format(chstr, resid))
         mutant = gml.ProteinMutant(orig.resname, target)
         atoms_add, hooks, geo_refs, bond_lengths, _, afters = mutant.atoms_to_add()
         atoms_remove = mutant.atoms_to_remove()
@@ -428,25 +536,24 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
                         break
             else:
                 aftnr = self.select_atom('{}resid {} and name {}'.format(chstr, resid, aft))
-            hindex = self.select_atom(hooksel)
             if len(geo_ref) == 2:
                 p1sel = '{}resid {} and name {}'.format(chstr, resid, geo_ref[0])
                 p2sel = '{}resid {} and name {}'.format(chstr, resid, geo_ref[1])
-                self.insert_atom(aftnr+1, self.atoms[hindex], atomsel=atomsel, hooksel=hooksel, bondlength=bond_length,
+                self.insert_atom(aftnr+1, self.get_atom(hooksel), atomsel=atomsel, hooksel=hooksel, bondlength=bond_length,
                                  p1_sel=p1sel, p2_sel=p2sel, atomname=atom_add)
             else:
                 vec = self._vector(geo_ref, resid, chain)
-                self.insert_atom(aftnr+1, self.atoms[hindex], atomsel=atomsel, hooksel=hooksel, bondlength=bond_length,
+                self.insert_atom(aftnr+1, self.get_atom(hooksel), atomsel=atomsel, hooksel=hooksel, bondlength=bond_length,
                                  vector=vec, atomname=atom_add)
 
-        for atom in self.select_atoms('{}resid {}'.format(chstr, resid)):
-            self.atoms[atom].resname = mutant.target_3l
+        for atom in self.get_atoms('{}resid {}'.format(chstr, resid)):
+            atom.resname = mutant.target_3l
         self.renumber_atoms()
 
     def _vector(self, atnames, resid, chain):
         chstr = 'chain {} and '.format(chain) if chain else ''
-        indices = [self.select_atom('{}resid {} and name {}'.format(chstr, resid, at)) for at in atnames]
-        vecs = [self._atoms_vec(self.atoms[indices[0]], self.atoms[q]) for q in indices[1:]]
+        atoms = [self.get_atom('{}resid {} and name {}'.format(chstr, resid, at)) for at in atnames]
+        vecs = [self._atoms_vec(atoms[0], at2) for at2 in atoms[1:]]
         nv = len(vecs)
         f = -1 if nv == 3 else 1
         return f*sum(v[0] for v in vecs)/nv, f*sum(v[1] for v in vecs)/nv, f*sum(v[2] for v in vecs)/nv
@@ -478,6 +585,14 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
         else:
             raise RuntimeError('Can\'t read box properties')
         return atoms, tuple(box), remarks
+
+    def add_conect(self, cutoff=1.55):
+        import numpy as np
+        coords = np.array(self.get_coords())
+        for n, atom in enumerate(self.atoms):
+            dists = np.linalg.norm(coords - np.array(atom.coords), axis=1)
+            selected = list(np.where(np.logical_and(dists<cutoff, dists>0.1))[0]+1)
+            self.conect[n+1] = selected
         
     def fill_beta(self, values, serials=None, smooth=False, ignore_mem=False):
         """
@@ -551,7 +666,19 @@ class Pdb:  # TODO optionally save as gro? & think of trajectories
                 outfile.write(line.strip() + '\n')
             for atom in self.atoms:
                 outfile.write(self._write_atom(atom))
+            for conect in self.conect.keys():
+                outfile.write(self._write_conect(conect, self.conect[conect]))
             outfile.write('ENDMDL\n')
+
+    def save_gro(self, outname='out.gro'):
+        if self.box[3] == self.box[4] == self.box[5] == 90.0:
+            with open(outname, 'w') as outfile:
+                outfile.write("written by gromologist\n{}\n".format(len(self.atoms)))
+                for atom in self.atoms:
+                    outfile.write(self._write_atom(atom, pdb=False))
+                outfile.write("{:10.5f}{:10.5f}{:10.5f}\n".format(*[x/10 for x in self.box[:3]]))
+        else:
+            raise RuntimeError("Only rectangular boxes are currently supported")
     
     def get_coords(self, subset=None):
         if subset:
@@ -596,7 +723,7 @@ class Atom:
         atomname = line[10:15].strip()
         atomnum = int(line[15:20].strip())
         x, y, z = [float(line[20+8*i:20+8*(i+1)].strip())*10 for i in range(3)]
-        return cls(data.format(atomnum, atomname, resname, resnum, x, y, z))
+        return cls(data.format(atomnum, atomname[:4], resname[:4], resnum, x, y, z))
 
     @classmethod
     def from_top_entry(cls, entry):
