@@ -74,7 +74,7 @@ class Section:
         """
         ssect = [s for s in self.subsections if s.header == section_name]
         if len(ssect) == 0:
-            raise KeyError
+            raise KeyError("Subsection {} not found, check your molecule topology!".format(section_name))
         elif len(ssect) > 1:
             raise RuntimeError("Error: subsection {} duplicated in {}".format(section_name, str(self)))
         return ssect[0]
@@ -87,8 +87,6 @@ class SectionMol(Section):
     """
     
     def __init__(self, content_list, top):
-        self.natoms = None
-        self.charge = None
         super().__init__(content_list, top)
         self.bonds = None
         self.mol_name = self.get_subsection('moleculetype').molname
@@ -101,6 +99,14 @@ class SectionMol(Section):
     def atoms(self):
         sub = self.get_subsection('atoms')
         return [entry for entry in sub if isinstance(entry, gml.EntryAtom)]
+
+    @property
+    def natoms(self):
+        return len(self.atoms)
+
+    @property
+    def charge(self):
+        return sum([a.charge for a in self.atoms])
     
     def select_atoms(self, selection_string):
         """
@@ -125,6 +131,12 @@ class SectionMol(Section):
             raise RuntimeError("Selection {} returned no atoms".format(selection_string, result))
         return result[0]
 
+    def get_atoms(self, selection_string):
+        return [self.atoms[i] for i in self.select_atoms(selection_string)]
+
+    def get_atom(self, selection_string):
+        return self.atoms[self.select_atom(selection_string)]
+
     def print_molecule(self):
         sub = self.get_subsection('atoms')
         for entry in sub:
@@ -137,7 +149,74 @@ class SectionMol(Section):
             if isinstance(ent, gml.EntryAtom) and ent.type_b is not None:
                 return True
         return False
-    
+
+    def _patch_alch(self):
+        self.update_dicts()
+        resnames = {a.resname for a in self.atoms}
+        if 'DTS' in resnames:
+            selected = [x+1 for x in self.select_atoms('resname DTS')]
+            sect = self.get_subsection('bonds')
+            for entry in sect:
+                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+                    entry.read_types()
+                    if entry.atom_names == ("C5'", "O5'") or entry.atom_names == ("O5'", "C5'"):
+                        if 'OS' in entry.types_state_a:
+                            entry.params_state_b = [0.40000, 0.0]
+                            entry.comment = entry.comment.rstrip() + ' fixed\n' if entry.comment else '; fixed\n'
+                        else:
+                            raise RuntimeError("Expected type OS in state A for bond C5'-O5' but found types {}, "
+                                               "aborting".format(' '.join(entry.types_state_a)))
+            sect = self.get_subsection('pairs')
+            to_remove = []
+            for n, entry in enumerate(sect):
+                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+                    entry.read_types()
+                    if entry.atom_names == ("P", "Ox5'") or entry.atom_names == ("P", "DO5'") or \
+                            entry.atom_names == ("O5'", "Hx5'") or entry.atom_names == ("O5'", "DH5'"):
+                        to_remove.append(n)
+            for en in to_remove[::-1]:
+                _ = sect.entries.pop(en)
+            atoms = sorted([x+1 for x in self.select_atoms("resname DTS and name O5' C5'")])
+            assert len(atoms) % 2 == 0
+            for pair in range(len(atoms)//2):
+                self.nullify_bonded(*atoms[2*pair:2*pair+2], 'angles')
+                self.nullify_bonded(*atoms[2 * pair:2 * pair + 2], 'dihedrals')
+        if 'DTD' in resnames or 'DTE' in resnames:
+            selected = [x+1 for x in self.select_atoms('resname DTD DTE')]
+            sect = self.get_subsection('bonds')
+            for entry in sect:
+                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+                    entry.read_types()
+                    if entry.types_state_a == ("DUM_CT", "DUM_C2") \
+                            and entry.atom_numbers[1] - entry.atom_numbers[0] > 30:
+                        entry.params_state_a[1] = 0.0
+                        entry.comment = entry.comment.rstrip() + ' fixed\n' if entry.comment else '; fixed\n'
+        if 'DTX' in resnames or 'DTY' in resnames:
+            selected = [x+1 for x in self.select_atoms('resname DTX DTY')]
+            sect = self.get_subsection('bonds')
+            for entry in sect:
+                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+                    entry.read_types()
+                    if entry.types_state_a == ("DUM_CT", "DUM_CT") \
+                            and entry.atom_numbers[1] - entry.atom_numbers[0] > 30:
+                        entry.params_state_a[1] = 0.0
+                        entry.comment = entry.comment.rstrip() + ' fixed\n' if entry.comment else '; fixed\n'
+
+    def nullify_bonded(self, atom1, atom2, subsection, stateB=True):
+        subs = self.get_subsection(subsection)
+        for entry in subs:
+            if isinstance(entry, gml.EntryBonded) and atom1 in entry.atom_numbers and atom2 in entry.atom_numbers:
+                entry.read_types()
+                if subsection == 'angles' and any(x.startswith('D') for x in entry.types_state_a):
+                    continue
+                if not entry.params_state_b:
+                    entry.params_state_b = entry.params_state_a[:]
+                if stateB:
+                    entry.params_state_b[1] = 0.0
+                else:
+                    entry.params_state_a[1] = 0.0
+                entry.comment = entry.comment.rstrip() + ' fixed\n' if entry.comment else '; fixed\n'
+
     def offset_numbering(self, offset, startfrom=0):
         """
         Offsets atom numbering starting from a specified position;
@@ -234,7 +313,7 @@ class SectionMol(Section):
                        entry.type_b[0] == "D" and entry.num in selected]
             print(dummies)
             while dummies:
-                to_remove = dummies[0]
+                to_remove = dummies[-1]
                 self.del_atom(to_remove.num)
                 dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type_b and
                            entry.type_b[0] == "D" and entry.num in selected]
@@ -325,11 +404,13 @@ class SectionMol(Section):
                         entry.types_state_b = None
         if remove_dummies:
             sub = self.get_subsection('atoms')
-            dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.atomname[0] == "D"]
+            dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type[0] == "D"
+                       and entry.num in selected]
             while dummies:
-                to_remove = dummies[0]
+                to_remove = dummies[-1]
                 self.del_atom(to_remove.num)
-                dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.atomname[0] == "D"]
+                dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type[0] == "D"
+                           and entry.num in selected]
         self.update_dicts()
     
     def add_atom(self, atom_number, atom_name, atom_type, charge=0.0, resid=None, resname=None, mass=None, prnt=True):
@@ -345,6 +426,7 @@ class SectionMol(Section):
         :param resid: int, residue number
         :param resname: str, residue name
         :param mass: float, mass of the atom
+        :param prnt: bool, whether to print the atom being added
         :return: None
         """
         subs_atoms = self.get_subsection('atoms')
@@ -367,7 +449,7 @@ class SectionMol(Section):
         elif resid and not resname:
             ref_entry = [e for e in atoms if (isinstance(e, gml.EntryAtom) and e.resid == resid)][0]
             resname = ref_entry.resname
-        if not mass:
+        if mass is None:
             param_sect = [s for s in self.top.sections if isinstance(s, SectionParam)][0]
             try:
                 param_entry = [e for e in param_sect.get_subsection('atomtypes').entries
@@ -406,7 +488,6 @@ class SectionMol(Section):
         self._del_atom(atom_number)
         self._del_params(atom_number)
         self.offset_numbering(-1, atom_number)
-        self.top.recalc_sys_params()
         self.update_dicts()
         if del_in_pdb:
             if self.top.pdb:
@@ -599,12 +680,58 @@ class SectionMol(Section):
         new_bond = [tuple(sorted([int(atom_own), int(atom_other)]))]
         new_angles = self._generate_angles(other, atom_own, atom_other)
         new_pairs, new_dihedrals = self._generate_14(other, atom_own, atom_other)
+        # TODO remove overlapping pairs between new_bond/new_angles and new_pairs for 4- and 5-membered rings
         for sub, entries in zip(['bonds', 'pairs', 'angles', 'dihedrals'],
                                 [new_bond, new_pairs, new_angles, new_dihedrals]):
             subsection = self.get_subsection(sub)
             subsection.add_entries([gml.EntryBonded(subsection.fstring.format(*entry, subsection.prmtype), subsection)
                                     for entry in entries])
 
+    def _remove_bond(self, at1, at2):
+        self._get_bonds()
+        bond_to_remove = [(at1, at2)]
+        if not (bond_to_remove[0] in self.bonds or tuple(x for x in bond_to_remove[0][::-1]) in self.bonds):
+            raise RuntimeError("Bond between atoms {} and {} not found in the topology".format(at1, at2))
+        angles_to_remove = self._generate_angles(self, at1, at2)
+        pairs_to_remove, dihedrals_to_remove = self._generate_14(self, at1, at2)
+        try:
+            impropers = self.get_subsection('impropers')
+        except KeyError:
+            impropers = self.get_subsection('dihedrals')
+        impropers_to_remove = []
+        for n, entry in enumerate(impropers.entries):
+            if isinstance(entry, gml.EntryBonded) and at1 in entry.atom_numbers and at2 in entry.atom_numbers:
+                impropers_to_remove.append(n)
+        for n in impropers_to_remove[::-1]:
+            _ = impropers.entries.pop(n)
+        try:
+            cmaps = self.get_subsection('cmap')
+        except KeyError:
+            pass
+        else:
+            cmap_to_remove = []
+            for n, entry in enumerate(cmaps.entries):
+                if isinstance(entry, gml.EntryBonded) and at1 in entry.atom_numbers and at2 in entry.atom_numbers:
+                    cmap_to_remove.append(n)
+            for n in cmap_to_remove[::-1]:
+                _ = cmaps.entries.pop(n)
+
+        def match(seq1, seqlist):
+            for seq2 in seqlist:
+                if all(i == j for i, j in zip(seq1, seq2)) or all(i == j for i, j in zip(seq1, seq2[::-1])):
+                    return True
+            return False
+
+        for sub, removable in zip(['bonds', 'pairs', 'angles', 'dihedrals'],
+                                  [bond_to_remove, pairs_to_remove, angles_to_remove, dihedrals_to_remove]):
+            subsection = self.get_subsection(sub)
+            to_remove = []
+            for n, e in enumerate(subsection.entries):
+                if isinstance(e, gml.EntryBonded) and match(e.atom_numbers, removable):
+                    to_remove.append(n)
+            for n in to_remove[::-1]:
+                _ = subsection.entries.pop(n)
+
     def _generate_angles(self, other, atom_own, atom_other):
         """
         Generates new angles when an additional bond is formed
@@ -681,11 +808,13 @@ class SectionMol(Section):
                     used_params.extend(ssub.find_used_ff_params())
         return used_params
 
-    def find_missing_ff_params(self, add_section='all', fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False):
+    def find_missing_ff_params(self, add_section='all', fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False,
+                               once=False):
         if add_section == 'all':  # TODO optionally add type/atomname labels in comment
             subsections_to_add = ['bonds', 'angles', 'dihedrals', 'impropers']
         else:
             subsections_to_add = [add_section]
+        self.printed = []
         for sub in subsections_to_add:
             try:
                 subsections = [s for s in self.subsections if s.header == sub]
@@ -694,7 +823,8 @@ class SectionMol(Section):
             else:
                 for ssub in subsections:
                     print(f"Searching in molecule {self.mol_name}, section {ssub}...")
-                    ssub.find_missing_ff_params(fix_by_analogy, fix_B_from_A, fix_A_from_B)
+                    ssub.find_missing_ff_params(fix_by_analogy, fix_B_from_A, fix_A_from_B, once)
+        del self.printed
 
     def label_types(self, add_section='all'):
         if add_section == 'all':
@@ -726,8 +856,9 @@ class SectionMol(Section):
                 rtp = glob(self.top.gromacs_dir + '/*ff/[am][em]*rtp')[rtpnum-1]
         elif self.top.rtp:
             rtp = ''
-        orig = self.atoms[self.select_atom('resid {} and name CA'.format(resid))]
+        orig = self.get_atom('resid {} and name CA'.format(resid))
         mutant = gml.ProteinMutant(orig.resname, target)
+        # TODO double-check [ pairs ]
         targ = mutant.target_3l
         print("\n  Mutating residue {} (resid {}) into {}\n".format(orig.resname, resid, targ))
         atoms_add, hooks, _, _, extra_bonds, afters = mutant.atoms_to_add()
@@ -737,9 +868,9 @@ class SectionMol(Section):
         if targ == 'HIS':
             targ = 'HSD' if ('HSD', 'CA') in types.keys() else 'HID'
         elif targ == 'GLY':
-            self.atoms[self.select_atom('resid {} and name HA'.format(resid))].atomname = 'HA1'
+            self.get_atom('resid {} and name HA'.format(resid)).atomname = 'HA1'
         if orig.resname == 'GLY':
-            self.atoms[self.select_atom('resid {} and name HA1'.format(resid))].atomname = 'HA'
+            self.get_atom('resid {} and name HA1'.format(resid)).atomname = 'HA'
         impropers_to_add = []
         impr_sub = self.get_subsection('impropers')
         atoms_sub = self.get_subsection('atoms')
@@ -749,10 +880,10 @@ class SectionMol(Section):
                            'CD': 'CD1', 'HD': 'HD1', 'HD1': 'HD11', 'HD2': 'HD12', 'HD3': 'HD13'}
             print("Removing atom {} from resid {} in topology".format(at, resid))
             try:
-                atnum = self.select_atom('resid {} and name {}'.format(resid, at))
+                atnum = self.get_atom('resid {} and name {}'.format(resid, at)).num
             except RuntimeError:
-                atnum = self.select_atom('resid {} and name {}'.format(resid, equivalents[at]))
-            self.del_atom(self.atoms[atnum].num, del_in_pdb=False)
+                atnum = self.get_atom('resid {} and name {}'.format(resid, equivalents[at])).num
+            self.del_atom(atnum, del_in_pdb=False)
         for atom_add, hook, aft in zip(atoms_add, hooks, afters):
             print("Adding atom {} to resid {} in topology".format(atom_add, resid))
             # if there are ambiguities in naming (two or more options):
@@ -805,27 +936,37 @@ class SectionMol(Section):
         for bond in extra_bonds:
             xsel = 'resid {} and name {}'.format(resid, bond[0])
             ysel = 'resid {} and name {}'.format(resid, bond[1])
-            xnum = self.atoms[self.select_atom(xsel)].num
-            ynum = self.atoms[self.select_atom(ysel)].num
+            xnum = self.get_atom(xsel).num
+            ynum = self.get_atom(ysel).num
             self.add_bond(xnum, ynum)
         atoms_sub.get_dicts(force_update=True)
         # looking for new impropers
         for imp in impropers_to_add:
             if set(imp).intersection(set(atoms_add)):
-                numbers = [atoms_sub.name_to_num[(resid, at)] for at in imp]
+                numbers = []
+                for at in imp:
+                    if at.startswith('-'):
+                        rid = resid - 1
+                        atx = at[1:]
+                    elif at.startswith('+'):
+                        rid = resid + 1
+                        atx = at[1:]
+                    else:
+                        rid = resid
+                        atx = at
+                    numbers.append(atoms_sub.name_to_num[(rid, atx)])
                 new_str = '{:5d} {:5d} {:5d} {:5d} {:>5s}\n'.format(*numbers, improper_type)
                 impr_sub.add_entry(gml.EntryBonded(new_str, impr_sub),
-                                   position=1 + [n for n, e in enumerate(impr_sub) if isinstance(e, gml.EntryBonded)][-1])
+                                   position=1+[n for n, e in enumerate(impr_sub) if isinstance(e, gml.EntryBonded)][-1])
         # repeating the mutation in the structure
         if mutate_in_pdb and self.top.pdb:
-            pdb_atoms = self._match_pdb_to_top(self.atoms[self.select_atom('resid {} and name CA'.format(resid))].num)
+            pdb_atoms = self._match_pdb_to_top(self.get_atom('resid {} and name CA'.format(resid)).num)
             pdb_chains = [self.top.pdb.atoms[at].chain for at in pdb_atoms]
             if len(pdb_atoms) == 1:
                 chain = '' if pdb_chains[0] == ' ' else pdb_chains[0]
                 self.top.pdb.mutate_protein_residue(resid, target, chain)
             elif len(pdb_atoms) > 1:
                 if any([pdb_chains[0] == pdb_chains[i] for i in range(1, len(pdb_chains))]):
-                    print()
                     response = input("The topology entry {} corresponds to multiple entries in the PDB; should we add "
                                      "chains to PDB and retry? (y/n)\n".format(self.mol_name))
                     if response.lower() == 'y':
@@ -853,7 +994,8 @@ class SectionMol(Section):
         reading_impropers = False
         reading_bondedtypes = False
         for line in rtp_cont:
-            if line.strip().startswith('[') and line.strip().split()[1] not in ['bondedtypes', 'atoms', 'bonds', 'impropers']:
+            if line.strip().startswith('[') and line.strip().split()[1] not in ['bondedtypes', 'atoms', 'bonds',
+                                                                                'impropers']:
                 resname = line.strip().split()[1]
             if line.strip().startswith('[') and line.strip().split()[1] == 'atoms':
                 reading_atoms = True
@@ -950,14 +1092,14 @@ class SectionMol(Section):
             else:
                 raise RuntimeError("{} is not a valid selection".format(sel))
         self.add_dummy_def(dummy_type)
-        orig = self.atoms[self.select_atoms('resid {}'.format(resid))[0]]
+        orig = self.get_atom('resid {}'.format(resid))
         ccharge = ccharge if ccharge is not None else round(orig.charge - 0.27, 4)
         atoms_add, hooks = [basename.replace('C', 'H') + str(i) for i in range(3)], 3 * [orig_name]
         for n, atom_add_hook in enumerate(zip(atoms_add, hooks), 1):
             atom_add, hook = atom_add_hook
             print("Adding atom {} to resid {} in the topology".format(atom_add, resid))
             hooksel = 'resid {} and name {}'.format(resid, orig_name)
-            hnum = self.atoms[self.select_atom(hooksel)].num
+            hnum = self.get_atom(hooksel).num
             atnum = hnum + n
             self.add_atom(atnum, atom_add, atom_type=dummy_type, charge=0, resid=resid, resname=orig.resname, mass=1.008)
             self.add_bond(hnum, atnum)
@@ -968,8 +1110,9 @@ class SectionMol(Section):
                 raise RuntimeError("Adding groups in PDB only supported for systems containing one molecule")
             bonds = self.list_bonds(returning=True)
             hook = [j for i in bonds for j in i if orig_name in i and orig_name != j][0]
+            print(hook)
             aligns = [j for i in bonds for j in i if hook in i and hook != j and orig_name != j]
-            aftnr = self.select_atom('resid {} and name {}'.format(resid, hook))
+            aftnr = self.select_atom('resid {} and name {}'.format(resid, orig_name))
             for n, aliat in enumerate(zip(aligns, atoms_add), 1):
                 ali, at = aliat
                 self.top.pdb.insert_atom(aftnr+n, self.top.pdb.atoms[aftnr],
@@ -1035,7 +1178,23 @@ class SectionParam(Section):
         """
         for sub in self.subsections:
             if 'dihedral' in sub.header:
-                sub.sort()
+                sub.sort()  # TODO if two have same periodicity & atoms, remove one with 0-s (PMX)
+
+    def find_used_ff_params(self, section='all'):
+        used_params = []
+        if section == 'all':  # TODO optionally add type/atomname labels in comment
+            subsections_to_add = ['atomtypes', 'pairtypes', 'nonbonded_params', 'constrainttypes']
+        else:
+            subsections_to_add = [section]
+        for sub in subsections_to_add:
+            try:
+                subsections = [s for s in self.subsections if s.header == sub]
+            except IndexError:
+                pass
+            else:
+                for ssub in subsections:
+                    used_params.extend(ssub.find_used_ff_params())
+        return used_params
 
     def clone_type(self, atomtype, prefix):
         """
@@ -1058,7 +1217,8 @@ class SectionParam(Section):
 
     def clean_unused(self, used_params, section='all'):
         matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
-                     'impropers': 'dihedraltypes'}
+                     'impropers': 'dihedraltypes', 'atomtypes': 'atomtypes', 'pairtypes': 'pairtypes',
+                     'nonbonded_params': 'nonbonded_params', 'constrainttypes': 'constrainttypes'}
         if section == 'all':
             subs = list(matchings.values())
         else:
@@ -1071,15 +1231,26 @@ class SectionParam(Section):
                     if not isinstance(entry, gml.EntryParam) or entry.identifier in used_params:
                         new_entries.append(entry)
                 ssect.entries = new_entries
+        atomtypes_used = {e.type for mol in self.top.molecules for e in mol.atoms
+                          if isinstance(e, gml.EntryAtom)}
+        atomtypes_b_used = {e.type_b for mol in self.top.molecules for e in mol.atoms
+                            if isinstance(e, gml.EntryAtom) and e.type_b}
+        atomtypes_used.union(atomtypes_b_used)
+        ssect = self.get_subsection('atomtypes')
+        new_entries = []
+        for entry in ssect.entries:
+            if not isinstance(entry, gml.EntryParam) or entry.types[0] in atomtypes_used:
+                new_entries.append(entry)
+        ssect.entries = new_entries
 
     def _remove_symm_dupl(self, prefix):
         for sub in self.subsections:
             if 'dihedral' in sub.header:
                 sub._remove_symm(prefix)
 
-    def get_opt_dih(self):
+    def get_opt_dih(self, types=False):
         ss = [sub for sub in self.subsections if sub.header == 'dihedraltypes' and int(sub.prmtype) == 9][0]
-        return ss.get_opt_dih()
+        return ss.get_opt_dih(types)
 
     def get_opt_dih_indices(self):
         ss = [sub for sub in self.subsections if sub.header == 'dihedraltypes' and int(sub.prmtype) == 9][0]
